// 
//  Copyright (C) 2008  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PegApp.hh"
#include "PegCmdButtons.hh"
#include "PegMenu.hh"
#include "PegParData.hh"
#include "PegParEntry.hh"
#include "PegParEntryString.hh"
#include "PegParEntryReal.hh"
#include "PegParEntryInteger.hh"
#include "PegParEntryBoolean.hh"
#include "PegPrefs.hh"
#include "PegPrefsDialog.hh"
#include "PegSaveChangesDialog.hh"
#include "PegStatusArea.hh"
#include "PegUtils.hh"

#include "CiaoGuiUtils.hh"
#include "CiaoGuiIPCServer.hh"
#include "CiaoGuiIPCMessage.hh"

#include "AhelpInterface.hh"

using namespace std;

PegParData    *thePData = NULL;

PegApp::PegApp()
{
   _init = 0; 
   _tips = gtk_tooltips_new();
   _prompts = gtk_tooltips_new();
   _prefs = new PegPrefs();
   _showtips = TRUE;
   _showprompts = FALSE; 
   reset_tips();
}

void PegApp::initialize(string toolfile, string xpaname)
{
   if (!_init)
   {
      _prefsDialog = NULL;
      _height = 400;
      _width  = 600; 
      _window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_default_size(GTK_WINDOW(_window), _width, _height);
      _vbox = gtk_vbox_new(false, 1);
      gtk_container_add (GTK_CONTAINER(_window), _vbox);

      gtk_container_set_border_width (GTK_CONTAINER (_window), 0);

      // handle exit via <alt><f4> or close button in titlebar 
      g_signal_connect(_window, "delete-event", G_CALLBACK(exit_cb), &_window);

      // create panes for contents and scrolled msg status area
      _vpane = gtk_vpaned_new ();
      gtk_box_pack_end(GTK_BOX(_vbox), _vpane, 1, 1, 0);

      // since panes can only support two entries, nest our panes. This is
      // a slight modification from old peg where the status message area 
      // was not paned. 
      _vpane2 = gtk_vpaned_new ();
      gtk_paned_pack1(GTK_PANED(_vpane), _vpane2, TRUE, TRUE);

      int loc = (int)(_height * 0.55); 
      gtk_paned_set_position(GTK_PANED(_vpane2), loc);

      // add scrolled window to display parameter widgets 
      _scrolled = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(_scrolled),
         GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      gtk_paned_pack1(GTK_PANED(_vpane2), _scrolled, TRUE, TRUE);

      // add status area
      _statusArea = new PegStatusArea(); 
      gtk_paned_pack2(GTK_PANED(_vpane), _statusArea->base_widget(), TRUE,TRUE);
      _statusArea->show();
      
      // add menu after status ara so analysis menu can output status
      _menu  = new PegMenu(toolfile, _tips); 
      gtk_box_pack_start(GTK_BOX(_vbox), _menu->base_widget(), 0, 0, 0);
      _menu->show(); 
      
      // command buttons require PegMenu for help menu 
      if ((_buttons = new PegCmdButtons()) != NULL)
         gtk_paned_pack2(GTK_PANED(_vpane2),_buttons->base_widget(),TRUE,TRUE);


      // position location of status msg area as last 20% of window
      loc = (int)(_height * 0.8); 
      gtk_paned_set_position(GTK_PANED(_vpane), loc);

      gtk_widget_show_all(_scrolled);
      gtk_widget_show_all(_vpane);
      gtk_widget_show_all(_window);

      GlobalAhelpInterface::SetDefaultContext("peg");  

      if (xpaname.length())  _xpaname = xpaname;
         else                _xpaname = "peg";

      add_ipc(); 

      _init = 1; 
  }
}

PegApp::~PegApp()
{
   cleanup();
}

int PegApp::exit_cb()
{
   theApp->exit_application(); 
   return 1;
}

int PegApp::save_changes()
{
   int rv = 0; 
   if (thePData && thePData->ChangesMade())
   {
      // changes were made so use see if the user wants to save first
      PegSaveChangesDialog *cmd = new PegSaveChangesDialog();
      rv = cmd->doit();
   }
   return rv; 
}

void PegApp::exit_application()
{
   if (save_changes() == 0)  theApp->cleanup();
}

int PegApp::SetFilename(string& filename, int argc, const char **argv, int status)
{
   _fname = filename;
   if (status && _fname.length())  theApp->status_message("Loading " + _fname);
   if (thePData) delete thePData;

   if ((thePData = new PegParData(_fname, argc, argv, _scrolled))!= NULL)
   {
      _autopath = 0;
      thePData->Setup();
      string fpath = thePData->GetFilePath();
      if (fpath.length() && fpath != _fname)
      {
	 _autopath = 1;
         theApp->status_message("Loaded " + _fname + " from " + fpath);
      }
   }
   return 0;
}

int PegApp::SetFilename(string& filename, int status)
{
   return SetFilename(filename, 0, NULL, status); 
}

int PegApp::SetFilename(char* fname, int argc, const char **argv, int status)
{
   if (fname)
   {
      string ff = fname;
      SetFilename(ff, argc, argv, status);
   }
   return 0;
}

int PegApp::SetFilename(char* fname, int status)
{
   if (fname)
   {
      string ff = fname;
      return SetFilename(ff, 0, NULL, status); 
   }
   return 1;
}

string PegApp::get_file_path()
{
   string fpath = "";
   if (thePData)  fpath = thePData->GetFilePath();
   return fpath; 
}

void PegApp::cleanup()
{
   // cleanup actually frees the resources and exits
   if (thePData)
   {
      delete thePData;
      thePData = NULL; 
   }
   if (_window)
   {
      gtk_widget_destroy (_window);
      _window = NULL;
   }
   if (_prefs)
   {
      delete _prefs; 
      _prefs = NULL; 
   }
   if (_prefsDialog)
   { 
      delete _prefsDialog;
      _prefsDialog = NULL;
   }
   exit(0); 
}

int PegApp::status_message(string val, int type)
{
   if (!_statusArea)  return -1;
   return _statusArea->add_message(val, type);
}

void PegApp::set_status_message(string val)
{
   theApp->status_message(val);
}

void PegApp::set_window_title(string val)
{
   if (_window)   gtk_window_set_title(GTK_WINDOW(_window), val.c_str());   
}

int PegApp::update_filelist(string fname)
{
   if (fname.length() && _menu)
   {
      return _menu->update_filelist(fname); 
   }
   return 0; 
}

int PegApp::save_file(string fname)
{
   if (!_buttons)  return 1; 
   bool overwrite = false;
   if (_prefs)  overwrite = _prefs->get_clobber(); 
   string fn = fname;
   if (fn.length() == 0 && !_autopath)  fn = get_file_path(); 
   return _buttons->save_file(fn, overwrite); 
}

int PegApp::get_preference(string pref, bool &val)
{
   if (!_prefs)  return 1;    // no prefs object
   return _prefs->get_preference(pref, val);
}

int PegApp::get_preference(string pref, string &val)
{
   if (!_prefs)  return 1;    // no prefs object
   return _prefs->get_preference(pref, val);
}

void PegApp::launch_preferences_dialog()
{
   if (!_prefs)
   {
      status_message("Unable to launch Preferences Dialog- no preferences object found.");
      return;
   }
   if (!_prefsDialog)
   {
      _prefsDialog = new PegPrefsDialog(_prefs, _tips);
      status_message("Preferences dialog will attempt to load $HOME/.pegrc");
   }
   if (_prefsDialog)    _prefsDialog->doit();
   reset_tips();
}


void PegApp::reset_tips()
{
   if (_tips)
   {
      if (_prefs) _showtips = _prefs->get_tooltips();
      if (_showtips)
         gtk_tooltips_enable(GTK_TOOLTIPS(_tips));
      else
         gtk_tooltips_disable(GTK_TOOLTIPS(_tips));
   }
   if (_prompts)
   {
      if (_prefs) _showprompts = _prefs->get_display_prompts();
      if (_showprompts)
         gtk_tooltips_enable(GTK_TOOLTIPS(_prompts));
      else
         gtk_tooltips_disable(GTK_TOOLTIPS(_prompts));
   }
   if (thePData)  thePData->ResetEntries();
}


void PegApp::enable_help_on_tool(bool val)
{
   if (_menu)  _menu->enable_help_current_tool(val); 
}


CiaoGuiIPCMessage* PegApp::FileMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;

   //  file [filename]
   int numParam = in_msg.GetTotalParams();
   if (numParam > 2)
   {
      // too many parameters
      returnMsg = new CiaoGuiIPCMessage("Error: Too many arguments... xpaset/get peg file [filename]\n");
   }
   else if (numParam == 2)
   {
      string param1;
      in_msg.GetParam(1, param1);
      if (_ciao_strcmp_cis(param1, "-path") == 0)
      {
         if (thePData) 
	    returnMsg = new CiaoGuiIPCMessage(thePData->GetFilePath());
      }
   }
   if (!returnMsg)
      returnMsg = new CiaoGuiIPCMessage(theApp->GetFilename());
   return returnMsg;
}

CiaoGuiIPCMessage* PegApp::SetMessageCB(CiaoGuiIPCMessage &in_msg)
{
   CiaoGuiIPCMessage* returnMsg = NULL;

   int numParam = in_msg.GetTotalParams();
   if (numParam < 3)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg set param value\n");

   string pname, val, tmpval;
   in_msg.GetParam(1, pname);
   in_msg.GetParam(2, val);

   // handle multiple entries (ie. filter syntax containing spaces 
   for (int rr = 3; rr < numParam; rr++)
   {
      in_msg.GetParam(rr, tmpval);
      val.append(" "); 
      val.append(tmpval); 
   }
   if (thePData) thePData->SetParamValue(pname, val); 
   return returnMsg;
}


CiaoGuiIPCMessage* PegApp::GetMessageCB(CiaoGuiIPCMessage &in_msg)
{
   if (in_msg.GetTotalParams() != 2)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg get param\n");

   string pname;
   in_msg.GetParam(1, pname);

   if (thePData)  pname = thePData->GetParamValue(pname); 
   
   return new CiaoGuiIPCMessage(pname);
}

CiaoGuiIPCMessage* PegApp::OpenMessageCB(CiaoGuiIPCMessage &in_msg)
{
   if (in_msg.GetTotalParams() != 2)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg open filename\n");
   string fname;
   in_msg.GetParam(1, fname);
   theApp->SetFilename(fname, 1);
   return NULL;
}

CiaoGuiIPCMessage* PegApp::PunlearnMessageCB(CiaoGuiIPCMessage &in_msg)
{
   if (in_msg.GetTotalParams() != 1)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg punlearn\n");
   PegCmdButtons *cmd_p = theApp->getCmdButtons();
   if (cmd_p)  cmd_p->punlearnCommand(); 
   return NULL;
}

CiaoGuiIPCMessage* PegApp::RunMessageCB(CiaoGuiIPCMessage &in_msg)
{
   if (in_msg.GetTotalParams() != 1)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg run\n");
   PegCmdButtons *cmd_p = theApp->getCmdButtons();
   if (cmd_p)  cmd_p->runCommand(); 
   return NULL;
}

CiaoGuiIPCMessage* PegApp::SaveMessageCB(CiaoGuiIPCMessage &in_msg)
{
   int cnt = in_msg.GetTotalParams(); 
   if (cnt > 2)
      return new CiaoGuiIPCMessage("Error: Incorrect arguments... xpaset/get peg save [filename]\n");
   PegCmdButtons *cmd_p = theApp->getCmdButtons();
   if (cmd_p)
   {
      string fname = "";
      if (cnt == 2) in_msg.GetParam(1, fname);
      cmd_p->save_file(fname);
   }
   return NULL;
}


void PegApp::add_ipc()
{
   ipc = new CiaoGuiIPCServer((char*)_xpaname.c_str(), 1);
   if (ipc)
   {
      ipc->AddCallback("file", &PegApp::FileMessageCB, 
         "file [-path]   : retrieves specified filename (or path to file)");
      ipc->AddCallback("get", &PegApp::GetMessageCB, 
         "get parameter  : retrieves the value of the specified parameter");
      ipc->AddCallback("open", &PegApp::OpenMessageCB, 
         "open filename  : opens the specified file");
      ipc->AddCallback("punlearn", &PegApp::PunlearnMessageCB, 
         "punlearn       : resets the parameter file to its system defaults");
      ipc->AddCallback("run", &PegApp::RunMessageCB, 
         "run            : runs the tool using the current gui parameter settings");
      ipc->AddCallback("save", &PegApp::SaveMessageCB, 
         "save [filename]: saves parameter values to specified (or current) file");
      ipc->AddCallback("set", &PegApp::SetMessageCB, 
         "set parameter value : sets parameter to value");
      ipc->GtkLoopAdd();
   }
}

