// 
//  Copyright (C) 2007-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismDataBlock.hh"
#include "PrismDataColumn.hh"
#include "PrismUtils.hh"
#include "PrismApp.hh"
#include "ascdm.h"

#include <iostream>
#include <sstream>
#include <float.h>
#include <string.h>
#include <stdlib.h>

#define PRISM_MAX_WRITE_SZ    5000
#define PRISM_MAX_NUM_EXPCOLS       10

using namespace std; 

PrismDataBlock::PrismDataBlock(dmBlock* block, bool filtered, int sz)
{
   _nrows = _ncols = _totcols  = _dispcols = 0;
   _cols = NULL;
   _keys = NULL; 
   _type    = dmUNKNOWNBLOCK;
   _filtered = filtered; 
   _maxExpandSz = sz; 
   if ((_block = block) != NULL) 
   {
      char tmp[dmMAXSTR+1];
      dmBlockGetName(_block,tmp,dmMAXSTR);
      _extnum = dmBlockGetNo(_block);
      _extname = tmp;
      _type = dmBlockGetType(_block);
      // for now, we're treating images and tables the same... 
      load_columns(); 
      load_header(); 
   }
}

PrismDataBlock::~PrismDataBlock()
{
   if (_block && !_filtered)   dmBlockClose(_block);
   if (_cols)
   {
      // remove old cols 
      delete[] _cols; 
      _cols = NULL; 
   }
   if (_keys && _nkeys > 0)
   {
     delete[] _keys;
     _keys = NULL; 
   }
}


int PrismDataBlock::load_columns()
{
   if (!_block)  return -1;  
   if (_type == dmTABLE)
   {
      dmDescriptor **dmColumns;
      long ii, vdim, adim, asize;
      int  ndx, vp; 
      dmColumns = dmTableOpenColumnList(_block, &_ncols);
      string *vcols = NULL;  
      long vcnt = 0; 

      // calcluate number of needed cols
      for (_dispcols = _totcols = ii = 0; ii < _ncols; ii++)
      {
         vdim = dmGetElementDim(dmColumns[ii]);
         adim = dmGetArrayDim(dmColumns[ii]);
         asize = dmGetArraySize(dmColumns[ii]);
	 if (dmGetDataType(dmColumns[ii]) == dmBIT)  asize *= 8; 
         _totcols += vdim;
         if (vdim > 1 | asize > 1)   _totcols++;  
      }
      _nrows = dmTableGetNoRows(_block);

      // load up column names to check for virtuals 
      if (_filtered)
      {
         char vbuff[1025];
         char vname[129];
         dmDataset *vset = dmBlockGetDataset(_block); 
         dmDatasetGetName(vset, vbuff, 1024);
         vset = dmDatasetOpen(vbuff);
         dmBlock *vblock = dmBlockOpen(vset, _extname.c_str()); 
         dmDescriptor **vdesc = dmTableOpenColumnList(vblock, &vcnt);
	 if (vcnt)
	 {
            vcols = new string[vcnt];
            for (long vv = 0; vv < vcnt; vv++)
            {
               dmGetName(vdesc[vv],vname,128);
               vcols[vv] = vname;
            }
         }
         dmBlockClose(vblock);
	 dmDatasetClose(vset); 
         free(vdesc); 
      }
      if (_cols)
      {
         // remove old cols 
	 delete[] _cols; 
         _cols = NULL; 
      }
      // setup column array with expanded entries
      if ((_cols = new PrismDataColumn[_totcols]) != NULL)
      {
         PrismDataColumn *tmp = &_cols[ndx=0];
	 string cname;
	 int ii, jj;
	 bool virt;        
         char aname[129];
         for (ii = 0; ii < _ncols; ii++)
         {
	    vdim = dmGetElementDim(dmColumns[ii]);
	    adim = dmGetArrayDim(dmColumns[ii]);
	    asize = dmGetArraySize(dmColumns[ii]);
            if (dmGetDataType(dmColumns[ii]) == dmBIT) asize *= 8; 

	    // flag virtual columns 
	    virt = 0;
            if (_filtered && vcols)
	    {
	       virt = 1; 
	       dmGetName(dmColumns[ii], aname, 128);
	       int vv = 0; 
	       while (vv < vcnt && virt) 
	          if (vcols[vv++] == aname)  virt = 0; 
	    }

	    if (vdim == 1)
	    {
	       if (adim == 0 && asize == 1)  // scalar 
	       {
		  tmp->set_desc(dmColumns[ii], ndx++);
		  if (virt)  tmp->set_virtual(true); 
	          tmp++; 
	       }
	       else   // array (1 or MD)  
	       {
		  vp = ndx++;
		  tmp->set_desc(dmColumns[ii], vp, -1, -1, -1, vp, vp+1,
		     vp+asize);
                  cname = tmp->get_name();
		  if (virt)  tmp->set_virtual(true); 
		  tmp++;
		  tmp->set_desc(dmColumns[ii],ndx++,-1,-1,-1,vp,vp+1,vp+asize);
		  tmp->set_name(cname);
		  tmp->set_component_type(PRISM_ARRAY_COMP); 
		  tmp->set_visible(0);
		  if (virt)  tmp->set_virtual(true); 
		  tmp++;
	       }
	    }
	    else
	    {
	       if (adim == 0 && asize == 1) // simple vector
	       {
		  vp = ndx++;
		  tmp->set_desc(dmColumns[ii], vp, vp, vp+1, vp+vdim);
		  cname = tmp->get_name(); 
		  if (virt)  tmp->set_virtual(true); 
	          tmp++; 
		  for (jj = 1; jj <= vdim; jj++)
		  {
		     tmp->set_desc(dmGetCpt(dmColumns[ii],jj),ndx++,vp,vp+1,vp+vdim);
		     tmp->set_visible(0);
		     tmp->set_vector_name(cname); 
		     tmp->set_component_type(PRISM_VECTOR_COMP); 
		     if (virt)  tmp->set_virtual(true); 
	             tmp++; 
	          }
	       }
	       else
	       {
		  vp = ndx++;
                  tmp->set_desc(dmColumns[ii], vp, vp, vp+1, vp+vdim*asize,
			        vp, vp+1, vp+vdim*asize);
                  cname = tmp->get_name();
		  if (virt)  tmp->set_virtual(true); 
	          tmp++; 
		  string tmpnam; 
	          for (int zz = 1; zz <= vdim; zz++)
                  {
                     tmp->set_desc(dmGetCpt(dmColumns[ii], zz), ndx++, vp,
			   vp+1, vp+vdim*asize, vp, vp+1, vp+vdim*asize);
                     tmpnam = tmp->get_name(); 
                     tmp->set_component_name(tmpnam); 
		     tmp->set_vector_name(cname); 
                     sprintf(aname,"%s(%s)",cname.c_str(), tmpnam.c_str());
                     tmp->set_name(aname);
		     tmp->set_component_type(PRISM_VECTOR_ARRAY_COMP); 
                     tmp->set_visible(0);
                     if (virt)  tmp->set_virtual(true); 
	             tmp++;
	          }
	       }
	    }
         }
      }
      if (vcols)    // clean up memory allocated for identifying virtual cols 
      {
         delete[] vcols; 
	 vcnt = 0;
         vcols = NULL; 
      }
   }
   else if (_type == dmIMAGE)
   {
      dmDescriptor *imdesc = dmImageGetDataDescriptor(_block); 
      if (_cols)
      {
         // remove old cols
         delete[] _cols;
         _cols = NULL;
         _dispcols = _ncols = _nrows = 0;
      }

      if ((_cols = new PrismDataColumn[_totcols=1]) != NULL) 
      {
         PrismDataColumn *tmp = _cols;
         tmp->set_image_desc(imdesc); 
         _nrows = tmp->get_imgrows(); 
         _dispcols = tmp->get_imgcols();
      }
   }

   for (int ii = 0; ii < _totcols; ii++)
       if (_cols[ii].get_type() == "double") 
          _cols[ii].set_format(theApp->get_data_format(_cols[ii].get_name(), _extname, 0));
       else if (_cols[ii].get_type() == "float") 
           _cols[ii].set_format(theApp->get_data_format(_cols[ii].get_name(), _extname, 1));
       else if (_cols[ii].get_type() == "long") 
           _cols[ii].set_format(theApp->get_data_format(_cols[ii].get_name(), _extname, 2));
       else if (_cols[ii].get_type() == "short") 
           _cols[ii].set_format(theApp->get_data_format(_cols[ii].get_name(), _extname, 3));
       else if (_cols[ii].get_type() == "string") 
           _cols[ii].set_format(theApp->get_data_format(_cols[ii].get_name(), _extname, 4));

   return 0;
}

string* PrismDataBlock::get_data(long colno, long start, long stop, long& cols_back)
{
   long sz, rr; 
   int image = is_image(); 
   char cbuffer[256];

   if (stop == 0)   stop = _nrows; 
   if (start <1)    return NULL;    // index starts at 1
   if ((sz = stop) < 1) return NULL;   // invalid index or no data
   if (colno < 0) return NULL; // invalid column  
   
   string* data;
   PrismDataColumn* col = NULL;
   
   if (image) 
   {
      long ll[4] = {colno+1, start, 1, 1};
      long ur[4] = {colno+1, stop, 1, 1};

      data = new string[sz]; 
      col = _cols;   // image stored as single col desc

      // scalar column- testing - move to column class 
      dmDescriptor* desc = col->get_descriptor();
      if (desc)
      {
	 dmDataType dtype = dmGetDataType(desc); 
         if (dtype == dmDOUBLE)
	 {
	    double *dd = new double[sz];
	    if (!dd) return NULL; 
	    dmImageDataGetSubArray_d(desc, ll, ur, dd); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), dd[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       ss.precision(16);
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << dd[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []dd; 
	 }
         else if (dtype == dmFLOAT)
	 {
	    float *ff = new float[sz];
	    if (!ff) return NULL; 
	    dmImageDataGetSubArray_f(desc, ll, ur, ff); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), ff[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       ss.precision(FLT_DIG);
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ff[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ff; 
	 }
         else if (dtype == dmLONG)
	 {
	    long *vv = new long[sz];
	    if (!vv) return NULL; 
	    dmImageDataGetSubArray_l(desc, ll, ur, vv); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), vv[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << vv[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []vv; 
	 }
         else if (dtype == dmULONG)
	 {
	    unsigned long *vv = new unsigned long[sz];
	    if (!vv) return NULL; 
	    dmImageDataGetSubArray_ul(desc, ll, ur, vv); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), vv[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << vv[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []vv; 
	 }
         else if (dtype == dmSHORT)
	 {
	    short *ii = new short int[sz];
	    if (!ii) return NULL; 
	    dmImageDataGetSubArray_s(desc, ll, ur, ii); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), ii[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ii[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ii; 
	 }
         else if (dtype == dmUSHORT)
	 {
	    unsigned short *ii = new unsigned short int[sz];
	    if (!ii) return NULL; 
	    dmImageDataGetSubArray_us(desc, ll, ur, ii); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), ii[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ii[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ii; 
	 }
         else if (dtype == dmBYTE)
	 {
	    unsigned char *uc = new unsigned char[sz];
	    if (!uc) return NULL; 
	    dmImageDataGetSubArray_ub(desc, ll, ur, uc); 
	    string format = col->get_format();
            if (format.length())
            {
               for (rr = 0; rr < sz; rr++)
               {
                  sprintf(cbuffer, format.c_str(), uc[rr]);
                  data[rr] = cbuffer;
               }
            }
            else
            {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << (int)uc[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []uc; 
	 }
      }
      cols_back = 1; 
      return data; 
   }

   // load table data 
   if (colno >= _totcols) return NULL; // invalid column  
   col = &_cols[colno]; 
   if (col->is_vector())
   {
      string txt = "(";
      int dm, vmin = col->vector_min(), vmax =col->vector_max();
      if (col->is_array())   // vector array
      {
         stringstream ss;
         data = new string[sz]; 

	 vmax = vmin + (dm = col->get_dim()) -1;
         for (rr = vmin; rr < vmax; rr++)
         {
            txt.append(_cols[rr].get_component_name());
            txt.append(",");
         }
         txt.append(_cols[vmax].get_component_name());
         char buffer[128];
	 if (dm)  vmax = (col->array_max() - col->array_min() - 1) / dm; 
         sprintf(buffer,")[0:%1d]", vmax);
         txt.append(buffer); 
	 for (rr = 0; rr < sz; rr++)   data[rr] = txt;
         cols_back = 1; 
      }
      else 
      {
         data = new string[sz]; 
         for (rr = vmin; rr < vmax; rr++)
         {
            txt.append(_cols[rr].get_name());
            txt.append(",");
         }
         txt.append(_cols[vmax].get_name());
         txt.append(")");
         for (rr = 0; rr < sz; rr++)   data[rr] = txt;
         cols_back = 1; 
      }
   }
   else if (col->is_array() || col->is_array_comp())
   {
      long asz = col->get_array_size(); 
      long acnt = asz * sz;
      stringstream ss;

      if (col->is_array_comp())
         data = new string[asz*sz]; 
      else
         data = new string[sz];   // array to big for inline expansion 
      char buffer[128];
      string cname = col->get_name();
      sprintf(buffer,"%s[0:%1ld]", cname.c_str(), asz-1);

      if (col->is_array_comp())
      {
         dmDescriptor* desc = col->get_descriptor();
         switch(dmGetDataType(desc))
         {
            case dmDOUBLE:
            {
               double* arr = new double[acnt];
               dmGetVectorArrays_d(desc,arr,1,asz, start, stop);
               long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          ss.precision(16);
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
               }
	       delete []arr; 
            }
            break; 
            case dmFLOAT:
            {
               float* arr = new float[acnt];
               dmGetVectorArrays_f(desc,arr,1,asz, start, stop);
               long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          ss.precision(8);
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmLONG:
            {
               long* arr = new long[acnt];
               dmGetVectorArrays_l(desc,arr,1,asz, start, stop);
               long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmULONG:
            {
               unsigned long* arr = new unsigned long[acnt];
               dmGetVectorArrays_ul(desc,arr,1,asz, start, stop);
               unsigned long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmSHORT:
            {
               short* arr = new short[acnt];
               dmGetVectorArrays_s(desc,arr,1,asz, start, stop);
               long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmUSHORT:
            {
               unsigned short* arr = new unsigned short[acnt];
               dmGetVectorArrays_us(desc,arr,1,asz, start, stop);
               unsigned long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ss << arr[rr] << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmBYTE:
            {
               unsigned char* arr = new unsigned char[acnt];
               dmGetVectorArrays_ub(desc,arr,1,asz, start, stop);
               long pp = 0; 
	       string format = col->get_format();
               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               {
                  unsigned int ui;
	          for (rr = 0; rr < acnt; rr++)
	          {
                     ui = arr[rr]; 
                     ss << ui << endl; 
                     ss >> data[pp++]; 
                     ss.clear();
	          }
	       }
	       delete []arr; 
            }
            break; 
            case dmBOOL:
            {
               dmBool* arr = new dmBool[acnt];
               dmGetVectorArrays_q(desc,arr,1,asz, start, stop);
               long pp = 0;

               for (rr = 0; rr < acnt; rr++)
               {
                  if (arr[rr] == dmTRUE)
                     data[pp++] = "TRUE";
                  else 
                     data[pp++] = "FALSE";
               }
               delete []arr;
            }
            break;
	    case dmBIT:
	    {
               unsigned char* arr = new unsigned char[acnt/8];
               dmGetVectorArrays_ub(desc,arr,1,asz/8, start, stop);
               long ndx, ps, pp = 0; 
	       for (rr = 0; rr < acnt; rr++)
	       {
	          ndx = rr / 8;
	          ps = rr % 8;
                  if (arr[ndx] & (0x80 >> ps))   data[pp++] = "1"; 
                     else                        data[pp++] = "0";
	       }
	       delete []arr; 
	    }
	    break; 
            case dmTEXT:
            {
               char** arr = new char*[acnt];
	       string format = col->get_format();
               long xx, pp = 0, blen = dmDescriptorGetLength(desc) + 1;
               for (xx = acnt; xx--; )   arr[xx] = new char[blen];
               dmGetVectorArrays_c(desc,arr,1,asz, start, stop);

               if (format.length())
               {
                  for (rr = 0; rr < acnt; rr++)
                  {
                     sprintf(cbuffer, format.c_str(), arr[rr]);
                     data[pp++] = cbuffer;
                  }
               }
               else
               { 
                  for (rr = 0; rr < acnt; rr++)
                  {
                     data[pp++] = arr[rr];
                  }
               }
               for (xx = acnt; xx--; )   delete [](arr[xx]);
               delete []arr;
            }
            break;
            default:
	    break; 
         }  // end switch 
         cols_back = asz + 1; 
      }
      else 
      {
	 // collapsed array data
	 for (rr = 0; rr < sz; rr++)  data[rr] = buffer; 
         cols_back = 1; 
      }
   }
   else
   {
      data = new string[sz]; 

      // scalar column- testing - move to column class 
      dmDescriptor* desc = col->get_descriptor();
      if (desc)
      {
	 char cbuffer[256]; 
	 dmDataType dtype = dmGetDataType(desc); 
         if (dtype == dmDOUBLE)
	 {
	    double *dd = new double[sz];
	    if (!dd) return NULL; 
	    dmGetScalars_d(desc, dd, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), dd[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       ss.precision(16);
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << dd[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []dd; 
	 }
         else if (dtype == dmFLOAT)
	 {
	    float *ff = new float[sz];
	    if (!ff) return NULL; 
	    dmGetScalars_f(desc, ff, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), ff[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       ss.precision(FLT_DIG);
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ff[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ff; 
	 }
         else if (dtype == dmLONG)
	 {
	    long *vv = new long[sz];
	    if (!vv) return NULL; 
	    dmGetScalars_l(desc, vv, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), vv[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << vv[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []vv; 
	 }
         else if (dtype == dmULONG)
	 {
	    unsigned long *vv = new unsigned long[sz];
	    if (!vv) return NULL; 
	    dmGetScalars_ul(desc, vv, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), vv[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << vv[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []vv; 
	 }
         else if (dtype == dmSHORT)
	 {
	    short *ii = new short int[sz];
	    if (!ii) return NULL; 
	    dmGetScalars_s(desc, ii, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), ii[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ii[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ii; 
	 }
         else if (dtype == dmUSHORT)
	 {
	    unsigned short *ii = new unsigned short int[sz];
	    if (!ii) return NULL; 
	    dmGetScalars_us(desc, ii, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), ii[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       for (rr = 0; rr < sz; rr++)
	       {
                  ss << ii[rr] << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []ii; 
	 }
         else if (dtype == dmBYTE)
	 {
	    unsigned char *uc = new unsigned char[sz];
	    if (!uc) return NULL; 
	    dmGetScalars_ub(desc, uc, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), uc[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       unsigned int ui; 
	       for (rr = sz; rr--; )
	       {
	          ui = uc[rr]; 
                  ss << ui << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []uc; 
	 }
         else if (dtype == dmTEXT)
	 {
	    long slen = dmDescriptorGetLength(desc) + 1;
            char **buff = new char*[sz];
	    for (rr = sz; rr--; ) buff[rr] = new char[slen];
	    if (!buff) return NULL;
	    dmGetScalars_c(desc, buff, slen, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), buff[rr]);
		  data[rr] = cbuffer; 
	          delete [](buff[rr]);
	       }
	    }
	    else 
	    {
	       for (rr = sz; rr--; ) 
	       {
	          data[rr] = buff[rr]; 
	          delete [](buff[rr]);
	       }
	    }
            delete []buff; 
	 }
         else if (dtype == dmBOOL)
         {
            int *ii = new int[sz];
            if (!ii) return NULL;
            dmGetScalars_q(desc, ii, start, stop);
            for (rr = 0; rr < sz; rr++)
            {
               if (ii[rr] == dmTRUE)
                  data[rr] = "TRUE";
               else
                  data[rr] = "FALSE";
            }
            delete []ii;
         }
         else if (dtype == dmBIT)
	 {
	    unsigned char *uc = new unsigned char[sz];
	    if (!uc) return NULL; 
	    dmGetScalars_bit(desc, uc, start, stop);
            string format = col->get_format();
            if (format.length())
            {
	       for (rr = 0; rr < sz; rr++)
	       {
	          sprintf(cbuffer, format.c_str(), uc[rr]);
		  data[rr] = cbuffer; 
	       }
	    }
	    else 
	    {
	       stringstream ss;
	       unsigned int ui; 
	       for (rr = sz; rr--; )
	       {
	          ui = uc[rr]; 
                  ss << ui << endl; 
                  ss >> data[rr]; 
	          ss.clear();
	       }
	    }
	    delete []uc; 
	 }
         cols_back = 1; 
      }
   }
   return data; 
}

int PrismDataBlock::load_header()
{
   if (!_block)  return -1;

   _nkeys = dmBlockGetNoKernelKeys(_block);
   if (_keys)  
   {
      delete []_keys; 
      _keys = NULL;
   }
   if (_nkeys == 0) return 0; 
   if ((_keys = new HeaderKey_t[_nkeys]) == NULL)  return 2;  // alloc failed

   char buf[2048], tt[256], kk[256];
   char *vp = NULL, *cp = NULL;

   for (int rr = 0; rr < _nkeys; rr++)
   {
      dmBlockGetKernelKey(_block, rr+1, buf, 2048);

      vp = strchr(buf,'=');
      *vp = '\0';
      vp +=2; 
      sscanf(buf, "Key %d: %s %s", &_keys[rr].pos, tt, kk);

      if (kk[0] == '*') _keys[rr].name = &kk[1]; 
      else              _keys[rr].name = kk;
      if (!strcmp(tt, "CMT"))      _keys[rr].type = "comment"; 
      else if (!strcmp(tt, "L"))   _keys[rr].type = "boolean"; 
      else if (!strcmp(tt, "I"))   _keys[rr].type = "long"; 
      else if (!strcmp(tt, "D"))   _keys[rr].type = "double"; 
      else if (!strcmp(tt, "R"))   _keys[rr].type = "float"; 
      else if (!strcmp(tt, "C"))   _keys[rr].type = "string"; 
      else _keys[rr].type = "unknown"; 
      if ((cp = strstr(vp, " / ")) != NULL)
      {
         *cp = '\0';
         cp += 3;
      }
      if (vp)  _keys[rr].value = vp; 
      if (cp)  _keys[rr].comment = cp; 
      dmDescriptor *tmpdesc = dmKeyOpen(_block, (char*)_keys[rr].name.c_str());
      if (tmpdesc) 
      {
	 dmGetUnit(tmpdesc, tt, 255);
         _keys[rr].units = tt;
         _keys[rr].edit = 1; 

	  string ss; 
	  dmDataType ty = dmGetDataType(tmpdesc);
	  if (!_dmdatatype_to_string(ty, ss))
	  {
	     _keys[rr].type = ss; 
	  }
      }
      else
      {
         _keys[rr].units = "";
         _keys[rr].edit = 0; 
      }
   }
   return 0; 
}


string* PrismDataBlock::get_image_data(long start, long stop)
{
   long sz, rr; 
   char cbuffer[256];

   if (!is_image())   return NULL;
   if (stop == 0)   stop = _nrows; 
   if (start <1)    return NULL;        // index starts at 1
   if ((sz = stop) < 1) return NULL;   // invalid index or no data
   if (_dispcols < 1)  return NULL;    // no image cols 
  
   sz *= _dispcols; 
   string * data = new string[sz]; 
   
   long ll[4] = {1, start, 1,1};
   long ur[4] = {_dispcols, start + stop -1, 1,1};

   // scalar column- testing - move to column class 
   dmDescriptor* desc = _cols[0].get_descriptor();
   string format = _cols[0].get_format();
   if (desc)
   {
      dmDataType dtype = dmGetDataType(desc); 
      if (dtype == dmDOUBLE)
      {
         double *dd = new double[sz];
         if (!dd) return NULL; 
         dmImageDataGetSubArray_d(desc, ll, ur, dd); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), dd[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            ss.precision(16);
            for (rr = 0; rr < sz; rr++)
            {
               ss << dd[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
	 }
         delete []dd; 
      }
      else if (dtype == dmFLOAT)
      {
         float *ff = new float[sz];
         if (!ff) return NULL; 
         dmImageDataGetSubArray_f(desc, ll, ur, ff); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), ff[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            ss.precision(FLT_DIG);
            for (rr = 0; rr < sz; rr++)
            {
               ss << ff[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
	 }
         delete []ff; 
      }
      else if (dtype == dmLONG)
      {
         long *vv = new long[sz];
         if (!vv) return NULL; 
         dmImageDataGetSubArray_l(desc, ll, ur, vv); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), vv[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            for (rr = 0; rr < sz; rr++)
            {
               ss << vv[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
         }
         delete []vv; 
      }
      else if (dtype == dmULONG)
      {
         unsigned long *vv = new unsigned long[sz];
         if (!vv) return NULL; 
         dmImageDataGetSubArray_ul(desc, ll, ur, vv); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), vv[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            for (rr = 0; rr < sz; rr++)
            {
               ss << vv[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
         }
         delete []vv; 
      }
      else if (dtype == dmSHORT)
      {
         short *ii = new short int[sz];
         if (!ii) return NULL; 
         dmImageDataGetSubArray_s(desc, ll, ur, ii); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), ii[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            for (rr = 0; rr < sz; rr++)
            {
               ss << ii[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
         }
         delete []ii; 
      }
      else if (dtype == dmUSHORT)
      {
         unsigned short *ii = new unsigned short int[sz];
         if (!ii) return NULL; 
         dmImageDataGetSubArray_us(desc, ll, ur, ii); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), ii[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            for (rr = 0; rr < sz; rr++)
            {
               ss << ii[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
         }
         delete []ii; 
      }
      else if (dtype == dmBYTE)
      {
         unsigned char *uc = new unsigned char[sz];
         if (!uc) return NULL; 
         dmImageDataGetSubArray_ub(desc, ll, ur, uc); 
	 if (format.length())
         {
	    for (rr = 0; rr < sz; rr++)
	    {
	       sprintf(cbuffer, format.c_str(), uc[rr]);
	       data[rr] = cbuffer;
	    }
	 }
	 else
	 {
            stringstream ss;
            for (rr = 0; rr < sz; rr++)
            {
               ss << (int)uc[rr] << endl; 
               ss >> data[rr]; 
               ss.clear();
            }
         }
         delete []uc; 
      }
   }
   return data; 
}

int PrismDataBlock::insert_keyword(string keyname, string value, string type, 
   string comment, string units)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified 
   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   if (desc)               return 3;   // key already exists 

   dmDataType dtyp; 
   if (_string_to_dmdatatype(type, dtyp) == 0)
   {
      switch (dtyp)
      {
         case dmTEXT:
            desc = dmKeyWrite_c(_block, (char*)keyname.c_str(),
	       (char*)value.c_str(), (char*)units.c_str(), (char*)comment.c_str());
	 break; 
         case dmLONG:
	 {
	    long val;
	    if (!value.length())
               desc = dmKeyWrite_l(_block, (char*)keyname.c_str(), 
	          0, (char*)units.c_str(), (char*)comment.c_str());
	    else if (_string_to_long(value, val) == 0)
               desc = dmKeyWrite_l(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 6;   // invalid long value 
	 }
	 break; 
         case dmULONG:
	 {
	    unsigned long val;
	    if (!value.length())
               desc = dmKeyWrite_ul(_block, (char*)keyname.c_str(), 
	          0, (char*)units.c_str(), (char*)comment.c_str());
	    else if (_string_to_ulong(value, val) == 0)
               desc = dmKeyWrite_ul(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 6;   // invalid unsigned long value 
	 }
	 break; 
         case dmDOUBLE:
	 {
	    double val = 0.0;
	    if (!value.length() || (_string_to_double(value, val) == 0))
               desc = dmKeyWrite_d(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 7;   // invalid double value 
	 }
	 break;
         case dmFLOAT:
	 {
	    float val;
	    if (!value.length())
               desc = dmKeyWrite_f(_block, (char*)keyname.c_str(), 
	          0.0, (char*)units.c_str(), (char*)comment.c_str());
	    else if (_string_to_float(value, val) == 0)
               desc = dmKeyWrite_f(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 8;   // invalid float value 
	 }
	 break; 
         case dmSHORT:
	 {
	    short val;
	    if (_string_to_short(value, val) == 0)
               desc = dmKeyWrite_s(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 9;   // invalid short value 
	 }
	 break; 
         case dmUSHORT:
	 {
	    unsigned short val;
	    if (_string_to_ushort(value, val) == 0)
               desc = dmKeyWrite_us(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 9;   // invalid short value 
	 }
	 break; 
         case dmBOOL:
	 {
	    bool val;
	    if (_string_to_bool(value, val) == 0)
               desc = dmKeyWrite_q(_block, (char*)keyname.c_str(), 
	          val, (char*)units.c_str(), (char*)comment.c_str());
	    else
	       return 10;   // invalid bool value 
	 }
	 break; 
         default:
            return 4;  // unsupported type 
	 break; 
      }
   }
   else
      return 4;  // unsupported type 
   return 0; 
}

int PrismDataBlock::delete_keyword(string keyname)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified

   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   if (!desc)              return 3;   // unable to find requested key

   return (dmDescriptorDelete(desc) != dmSUCCESS);
}

int PrismDataBlock::change_keyword_value(string keyname, string value)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified 

   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   
   if (!desc)              return 3;   // unable to find requested key
   
   dmDataType typ = dmGetDataType(desc);
   switch (typ)
   {
      case dmTEXT:
	 dmSetScalar_c(desc, (char*)value.c_str());
      break;
      case dmDOUBLE:
      {
	 double vv;
	 if (_string_to_double(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_d(desc, vv);
      }
      break; 
      case dmFLOAT:
      {
	 float vv;
	 if (_string_to_float(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_f(desc, vv);
      }
      break; 
      case dmLONG:
      {
	 long vv; 
	 if (_string_to_long(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_l(desc, vv);
      }
      break; 
      case dmULONG:
      {
	 unsigned long vv; 
	 if (_string_to_ulong(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_ul(desc, vv);
      }
      break; 
      case dmSHORT:
      {
	 short vv; 
	 if (_string_to_short(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_s(desc, vv);
      }
      break; 
      case dmUSHORT:
      {
	 unsigned short vv; 
	 if (_string_to_ushort(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_us(desc, vv);
      }
      break; 
      case dmBOOL:
      {
	 bool vv;
	 if (_string_to_bool(value, vv)) return 1; // exit if conv failed 
	 dmSetScalar_q(desc, vv);
      }
      break; 
      default:
      {
         cout << " keyword conversion not supported for type " << typ << endl; 
      }
   }
   return 0; 
}

int PrismDataBlock::change_keyword_name(string keyname, string value)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified 
   if (!value.length())    return 5;   // no keyname specified 

   dmDescriptor* desc2 = dmKeyOpen(_block, (char*)value.c_str());
   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   
   if (!desc)              return 3;   // unable to find requested key
   if (desc2)              return 4;   // key with name already exists 

   dmSetName(desc, (char*)value.c_str()); 
   
   return 0; 
}

int PrismDataBlock::change_keyword_comment(string keyname, string value)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified 

   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   
   if (!desc)              return 3;   // unable to find requested key

   dmSetDesc(desc, (char*)value.c_str()); 
   
   return 0; 
}

int PrismDataBlock::change_keyword_units(string keyname, string value)
{
   if (!_block)            return 1;   // block not open / DNE
   if (!keyname.length())  return 2;   // no keyname specified 

   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   
   if (!desc)              return 3;   // unable to find requested key

   dmSetUnit(desc, (char*)value.c_str()); 
   
   return 0; 
}

bool PrismDataBlock::key_exists(string keyname)
{
   if (!_block)   return false; 
   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   return (desc != NULL);
}

int PrismDataBlock::get_key_fields(string keyname, string &value, string &comment)
{
   if (!_block)   return 1;
   if (!keyname.length())    return 2; 
   dmDescriptor* desc = dmKeyOpen(_block, (char*)keyname.c_str());
   if (!desc)     return 3; 

   char buf[256];
   dmGetDesc(desc, buf, 255);
   comment = buf; 
   stringstream ss;

   switch(dmGetDataType(desc)) 
   {
      case dmTEXT:
      {
         dmGetScalar_c(desc, buf, 255);
	 value = buf;
      }
      break; 
      case dmDOUBLE:
      {
         ss << dmGetScalar_d(desc);
	 ss >> value; 
      }
      break; 
      case dmFLOAT:
      {
         ss << dmGetScalar_f(desc);
	 ss >> value; 
      }
      break; 
      case dmLONG:
      {
         ss << dmGetScalar_l(desc);
	 ss >> value; 
      }
      break; 
      case dmULONG:
      {
         ss << dmGetScalar_ul(desc);
	 ss >> value; 
      }
      break; 
      case dmSHORT:
      {
         ss << dmGetScalar_s(desc);
	 ss >> value; 
      }
      break; 
      case dmUSHORT:
      {
         ss << dmGetScalar_us(desc);
	 ss >> value; 
      }
      break; 
      case dmBOOL:
      {
         ss << dmGetScalar_q(desc);
	 ss >> value; 
      }
      break; 
         return 4;    // unsupported type
      default:
      break; 
   }
   return 0; 
}

int PrismDataBlock::validate_value(int col, long row, string value)
{
   if (row < 1)        return 1;   // invalid row number (dm starts @ row 1) 
   if (row > _nrows)   return 2;   // specified row DNE  
   if (col < 0)        return 3;   // invalid col number (we index cols from 0)
   if (!_cols)         return 4;   // no columns to edit 
   PrismDataColumn* cp; 

   if (_type != dmIMAGE)
   {
      if (col > _totcols) return 5;   // specified col DNE
      if (dmTableSetRow(_block, row) == dmNOMOREROWS) return 7; // invalid row 
      cp = &_cols[col]; 
   }
   else
   {
      if (col > _dispcols) return 5;   // specified col DNE
      cp = _cols; 
   }

   int rv = cp->validate_value(value);

   if (rv) _errmsg = cp->get_errmsg(); 
   else    _errmsg = "";
   return rv;
}

int PrismDataBlock::set_value(int col, long row, string value, long arrpos)
{
   if (row < 1)        return 1;   // invalid row number (dm starts @ row 1) 
   if (row > _nrows)   return 2;   // specified row DNE  
   if (col < 0)        return 3;   // invalid col number (we index cols from 0)
   if (!_cols)         return 4;   // no columns to edit 
   int rv = 0; 

   if (_type != dmIMAGE)
   {
      if (col >= _totcols) return 5;   // specified col DNE
      if (dmTableSetRow(_block, row) == dmNOMOREROWS) return 7; // invalid row 
      if (arrpos > -1 && _cols[col].is_array())  
         rv = (_cols[col]).set_array_value(arrpos, value); // set post-expanded array 
      else  
         rv = (_cols[col]).set_value(value); 
   }
   else
   {
      if (col >= _dispcols) return 5;   // specified col DNE
      // set image value
      rv = (_cols[0]).set_image_value(arrpos,row,value); 
   }
   if (rv == 0  && theApp) theApp->set_modified();
   return rv; 
}

int PrismDataBlock::append_row(long copyrow)
{
   if (_type == dmIMAGE)  return 1;   // can't append row to image 
   if (copyrow != dmENDOFTABLE && (copyrow < 0 || copyrow > _nrows))  
      return 2;   // invalid copy row specified 

   void *blk_p = dmTableAllocRow(_block);
   dmTableSetRow(_block, copyrow);
   int val = dmTableGetRowSize(_block);

   // copy the specified row or zero out the new row 
   if (copyrow == 0)
      memset(blk_p, 0, val);
   else 
      dmTableGetRow(_block, blk_p); 

   // move to end of table 
   dmTableSetRow(_block, dmENDOFTABLE);
   val = (dmTablePutRow(_block, blk_p) == dmBADROW);

   if (val != 0)  _nrows++;
      
   return val; 
}


int PrismDataBlock::append_column(PrismColumnInfra_p colstruct)
{
   if (_type == dmIMAGE)  return 1;   // can't append column to image
   if (!colstruct)        return 2;   // don't have column info  
   long ssz = colstruct->ssize;
   dmDataType type; 
   dmDescriptor * desc = NULL; 

   switch(colstruct->dtype)
   {
      case PRISM_SHORT_TYPE:     type = dmSHORT;   break;  
      case PRISM_LONG_TYPE:      type = dmLONG;    break;  
      case PRISM_FLOAT_TYPE:     type = dmFLOAT;   break;  
      case PRISM_DOUBLE_TYPE:    type = dmDOUBLE;  break;  
      case PRISM_STRING_TYPE:    type = dmTEXT;    break;  
      case PRISM_UBYTE_TYPE:     type = dmBYTE;    break;  
      case PRISM_ULONG_TYPE:     type = dmULONG;   break;
      case PRISM_USHORT_TYPE:    type = dmUSHORT;  break;
      case PRISM_BOOL_TYPE:	 type = dmBOOL;    break; 
      default:
         return 3;      // unsupported data type 
      break; 
   }

   switch(colstruct->ctype)
   {
      case PRISM_SCALAR_COL:
      {
         desc = dmColumnCreate(_block, (char*)colstruct->name.c_str(), type, 
	                       ssz, (char*)colstruct->units.c_str(),
			       (char*)colstruct->comment.c_str()); 
         if (colstruct->values)
	 {
            long vv, vpps, vcnt;

            if (_nrows < PRISM_MAX_WRITE_SZ)  vcnt = _nrows;
               else vcnt = PRISM_MAX_WRITE_SZ;
            vpps = _nrows / PRISM_MAX_WRITE_SZ;

	    switch (type)
	    {
	       case dmSHORT:
	       {
	          short *varr, val;
		  if (!_string_to_short(colstruct->values[0], val))
		  {
		     varr = new short[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_s(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_s(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       case dmUSHORT:
	       {
	          unsigned short *varr, val;
		  if (!_string_to_ushort(colstruct->values[0], val))
		  {
		     varr = new unsigned short[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_us(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_us(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       case dmLONG:
	       {
	          long *varr, val;
		  if (!_string_to_long(colstruct->values[0], val))
		  {
		     varr = new long[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_l(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_l(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       case dmULONG:
	       {
	          unsigned long *varr, val;
		  if (!_string_to_ulong(colstruct->values[0], val))
		  {
		     varr = new unsigned long[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_ul(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_ul(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       case dmDOUBLE:
	       {
	          double *varr, val;
		  if (!_string_to_double(colstruct->values[0], val))
		  {
		     varr = new double[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_d(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_d(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       case dmFLOAT:
	       {
	          float *varr, val;
		  if (!_string_to_float(colstruct->values[0], val))
		  {
		     varr = new float[vcnt];
		     for (vv = vcnt; vv--; )  varr[vv] = val; 
		     for(vv = 0; vv < vpps; vv++)
  	                dmSetScalars_f(desc, varr, vv*PRISM_MAX_WRITE_SZ+1,
			               PRISM_MAX_WRITE_SZ); 
		     dmSetScalars_f(desc, varr, vpps*PRISM_MAX_WRITE_SZ+1, 
		                    _nrows - vpps*PRISM_MAX_WRITE_SZ);
		     delete[] varr; 
                  } 
	       }
	       break;
	       default:
	          return 6; // unknown or unsupported type 
	       break; 
	    }
	 }
      }
      break;
      case PRISM_ARRAY_COL:
      {
	 long asize = colstruct->asize; 
         desc = dmColumnCreateArray(_block, (char*)colstruct->name.c_str(),
	                            type, ssz, (char*)colstruct->units.c_str(), 
	                            (char*)colstruct->comment.c_str(), asize); 

         if (colstruct->values && colstruct->valcnt == asize)
	 {
            long vv, vpps, vcnt;

            if (_nrows < PRISM_MAX_WRITE_SZ)  vcnt = _nrows;
               else vcnt = PRISM_MAX_WRITE_SZ;
            vpps = _nrows / PRISM_MAX_WRITE_SZ;

	    switch (type)
	    {
	       case dmSHORT:
	       {
	          short *varr, sts = 0;
		  varr = new short[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_short(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_s(desc, varr, 1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_s(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmUSHORT:
	       {
	          unsigned short *varr, sts = 0;
		  varr = new unsigned short[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_ushort(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_us(desc,varr,1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_us(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmLONG:
	       {
	          long *varr, sts = 0;
		  varr = new long[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_long(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_l(desc, varr, 1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_l(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmULONG:
	       {
	          unsigned long *varr, sts = 0;
		  varr = new unsigned long[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_ulong(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_ul(desc,varr, 1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_ul(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmFLOAT:
	       {
	          float *varr, sts = 0;
		  varr = new float[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_float(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_f(desc, varr, 1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_f(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmDOUBLE:
	       {
	          double *varr, sts = 0;
		  varr = new double[vcnt * asize]; 

		  for (vv = asize; vv--; )  // populate initial vector array
		     sts += _string_to_double(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = asize; cc--; )  varr[vv*asize+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectorArrays_d(desc, varr, 1, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectorArrays_d(desc, varr, 1, vpps*PRISM_MAX_WRITE_SZ+1,
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       default:
	          return 6; // unknown or unsupported type 
	       break; 
            } // end switch
	 } 
      }
      break;
      case PRISM_VECTOR_COL:
      {
	 long vdim = (long)colstruct->vsize;

         if (vdim < 2)   return 4;  // vector needs at least 2 components 
         if (colstruct->components == NULL)  return 5; // component names missing

         char** cptnames = new char*[vdim];
         string* tmp = colstruct->components; 
         for (int rr =0; rr < vdim; rr++)
	    if (tmp != NULL) 
	    {
	       cptnames[rr] = (char*)tmp->c_str();
	       tmp++;
	    }
         desc = dmColumnCreateVector(_block, (char*)colstruct->name.c_str(),
	                            type, ssz, (char*)colstruct->units.c_str(),
				     (char*)colstruct->comment.c_str(), 
				     cptnames, vdim);
         delete[] cptnames; 

         if (colstruct->values && colstruct->valcnt == colstruct->vsize)
	 {
            long vv, vpps, vcnt;

            if (_nrows < PRISM_MAX_WRITE_SZ)  vcnt = _nrows;
               else vcnt = PRISM_MAX_WRITE_SZ;
            vpps = _nrows / PRISM_MAX_WRITE_SZ;

	    switch (type)
	    {
	       case dmSHORT:
	       {
	          short *varr, sts = 0;
		  varr = new short[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_short(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_s(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_s(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmUSHORT:
	       {
	          unsigned short *varr, sts = 0;
		  varr = new unsigned short[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_ushort(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_us(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_us(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmLONG:
	       {
	          long *varr, sts = 0;
		  varr = new long[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_long(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_l(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_l(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmULONG:
	       {
	          unsigned long *varr, sts = 0;
		  varr = new unsigned long[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_ulong(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_ul(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_ul(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmFLOAT:
	       {
	          float *varr, sts = 0;
		  varr = new float[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_float(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_f(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_f(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       case dmDOUBLE:
	       {
	          double *varr, sts = 0;
		  varr = new double[vcnt * vdim]; 

		  for (vv = vdim; vv--; )  // populate initial vector array
		     sts += _string_to_double(colstruct->values[vv], varr[vv]);
                  if (sts) 
		  {
		     delete[] varr;
		     return 5;  // unable to use initial values 
		  }
		  for (vv = 1; vv < vcnt; vv++)  // populate entire buffer
		     for (int cc = vdim; cc--; )  varr[vv*vdim+cc] = varr[cc];
		  for(vv = 0; vv < vpps; vv++)
  	             dmSetVectors_d(desc, varr, vdim, vv*PRISM_MAX_WRITE_SZ+1,
		             PRISM_MAX_WRITE_SZ); 
		  dmSetVectors_d(desc, varr, vdim, vpps*PRISM_MAX_WRITE_SZ+1, 
		                 _nrows - vpps*PRISM_MAX_WRITE_SZ);
		  delete[] varr; 
	       }
	       break;
	       default:
	          return 6; // unknown or unsupported type 
	       break; 
            } // end switch
         }  // end if (colstruct->values)
      }
      break; 
      default:
         return 4;    // unknown/unsupported type of column 
      break; 
   }
   return (desc == NULL); 
}


int PrismDataBlock::get_column_names(string* &colnames)
{
   int ii, jj = 0, cnt = 0;
   if (!_cols)  return 0;

   PrismDataColumn *tmp = _cols;
   for (ii = 0; ii < _totcols; ii++)
   {
      if (!tmp->is_array_comp())  cnt++;  
      tmp++;
   }
   if (!cnt) return 0;

   colnames = new string[cnt];
   tmp = _cols;
   for (ii = 0; ii < _totcols; ii++)
   {
      if (!tmp->is_array_comp())   colnames[jj++] = tmp->get_name(); 
      tmp++; 
   }
   return cnt; 
}

void PrismDataBlock::flush_changes()
{
   if (_block)   dmBlockFlush(_block);
}

