// 
//  Copyright (C) 2007-2008,2013  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "PrismDataSet.hh"
#include "PrismBackupDialog.hh"
#include "PrismDataBlock.hh"
#include "PrismUtils.hh"
#include "PrismApp.hh"
#include <iostream>
#include <sstream> 

using namespace std; 

PrismDataSet::PrismDataSet(string filename)
{
   _dataset = NULL; 
   _nblocks = 0;
   _currBlock = -1;      // since blocks are numbered 0..n-1 
   _status = 1;          // 0 = good, non-0 = bad (init to 1 - no file loaded)  
   _fname = filename;
   _block = NULL; 
   _fblock = NULL; 
   _filtblock = -1; 
   _editmode = 0;
   _modified = 0; 
   _ascii = false; 

   if (theApp)
   {
      if (theApp->get_preference("edit.confirm.dialog", _firstunlock))
         _firstunlock = true;
   }
   else
      _firstunlock = true; 
   _extensions = NULL; 
   open_dataset();

}

PrismDataSet::~PrismDataSet()
{
   if (_block)
   {
      delete _block;
      _block = NULL;
   }
   if (_dataset)  
   {
      dmDatasetClose(_dataset);
      _dataset = NULL;
   }
   if (_extensions) 
   {
      delete []_extensions;
      _extensions = NULL; 
   }
}


int PrismDataSet::open_dataset()
{
   if (!_fname.length())
      return _status = 2;    // no file specified
   else if (dmDatasetAccess(_fname.c_str(), "R") != dmTRUE)
      return _status = 3;    // unable to access file 

  if (_file_writable(_fname)) 
     _editmode = 1;   // wread-write
  else 
     _editmode = -1;  // read-only file


   int pos = _fname.find("[");
   if (pos != -1)
   {
      _filter = _fname.substr(pos, _fname.size());
      _fblock;  

      if (!(_fblock = dmTableOpen((char*)_fname.c_str())))
          if (!(_fblock = dmImageOpen((char*)_fname.c_str())))
             return _status = 4; // filter does not exist/ not supported 
      _dataset = dmBlockGetDataset(_fblock);
      _currBlock = _filtblock = dmBlockGetNo(_fblock); 
   }
   else
   {
      _fblock = NULL; 
      if ((_dataset = dmDatasetOpenUpdate(_fname.c_str())) == NULL)
         return _status = 4;    // dataset open failed 
   }

   _status = 0; 

   _nblocks = dmDatasetGetNoBlocks(_dataset);

   // get extension names
   browse_extensions();

   // if data had a filter, use that block, otherwise open the first extension.
   if (pos != -1)
      _status = open_extension(_filtblock);
   else
   {
     _status = open_extension(1);
     if (_block && _block->is_image() && !_block->get_nrows() && 
         !_block->get_ncols() && (_nblocks > 1))  
     {
        // if first extension is null header open next block 
        _status = open_extension(2); 
     }
   }
   
   // determine whether this is as ascii/text file or not
   if (_dataset)
   {
       char kern[64];
       dmDatasetGetKernel(_dataset, kern, 64);
       _ascii =  (strcmp(kern, "TEXT") == 0); 
   }

   return _status; 
}


int PrismDataSet::open_extension(int posn)
{
   _block = NULL; 
   if (_status)  return 5; // no data set  
   if ((posn < 1) || (posn > _nblocks + 1))   return 6;    // invalid ext number 
   if (_fblock && _filtblock == posn)
   {
      // work around since dmDatasetMoveToBlock loses filter info when moving 
      _currBlock = posn;
      _block = new PrismDataBlock(_fblock, 1); 
   }
   else
   {
      dmBlock *tmpBlock;
      if ((tmpBlock = dmDatasetMoveToBlock(_dataset, posn)) != NULL)
      {
         _currBlock = posn; 

         dmBlockType tmpType = dmBlockGetType(tmpBlock);
         if ( tmpType == dmTABLE )
         {
            _block = new PrismDataBlock(tmpBlock);
         }
         else if (tmpType == dmIMAGE)
         {
	    // for now handle the same
            _block = new PrismDataBlock(tmpBlock);
         }
      }
   }
   return 0; 
}


int PrismDataSet::open_extension(string ename)
{
   int filtered = 0; 
   if (_status)  return 5; // no data set  

   // close any currently open block
   if (_block)
   {
      delete _block;
      _block = NULL;
   }

   dmBlock *tmpBlock = dmBlockOpen(_dataset, ename.c_str());
   if (tmpBlock == NULL)  return 7;   // block name DNE

   if (dmBlockGetNo(tmpBlock) == _filtblock)
   {
      dmBlockClose(tmpBlock);
      tmpBlock = _fblock;
      filtered = 1;
   }
   

   dmBlockType tmpType = dmBlockGetType(tmpBlock);
   if ( tmpType == dmTABLE )
   {
      _block = new PrismDataBlock(tmpBlock, filtered);
      _currBlock = dmBlockGetNo(tmpBlock);
   }
   else if (tmpType == dmIMAGE)
   {
      // for now handle the same
     _block = new PrismDataBlock(tmpBlock, filtered);
      _currBlock = dmBlockGetNo(tmpBlock);
   }
   return 0; 
}


long PrismDataSet::get_nrows()
{
   if (_block)  return _block->get_nrows(); 
   return 0; 
}


int PrismDataSet::get_ncols()
{
   if (_block)  return _block->get_ncols(); 
   return 0; 
}

int PrismDataSet::get_extensions(list<ExtensionInfo_t> &ext_list)
{
   for (int rr = 0; rr < _nblocks; rr++)
      ext_list.push_back(_extensions[rr]);  
   return 0; 
}


int PrismDataSet::browse_extensions()
{
   if ((_extensions = new ExtensionInfo_t[_nblocks]) != NULL)
   {
      int open_block = _currBlock;
      ExtensionInfo_p ext_p = _extensions; 
      dmBlock* block; 
      char blName[1025], dim[80];
      long nrows, ncols;
      for (int rr = 1; rr < _nblocks+1; rr++)
      {
	 // workaround- dmDatasetMoveToBlock loses filter on move 
	 if (rr == _filtblock)
	    block = _fblock;
	 else 
            block = dmDatasetMoveToBlock(_dataset, rr);  
         dmBlockGetName(block, blName, 1024);
         ext_p->name = blName; 
         if (dmBlockGetType(block) == dmTABLE)
         {
            ext_p->type = "table";
            ncols = dmTableGetNoCols(block);
            nrows = dmTableGetNoRows(block);
	    sprintf(dim, "%1ld %s, %1ld %s", 
	       ncols, (ncols == 1)? "col":"cols",  
	       nrows, (nrows == 1)? "row":"rows");
	    ext_p->dim = dim; 
         }
         else
         {
	    stringstream ss; 
            ext_p->type = "image";
            dmDescriptor *imdesc = dmImageGetDataDescriptor(block);
	    long *axisLengths = 0;
            long naxes = dmGetArrayDimensions(imdesc,&axisLengths);
	    if (naxes < 1)
	        ext_p->dim = "NULL";
	    else 
	    {
               ss << axisLengths[0];
	       for (int zz = 1; zz < naxes; zz++)
	       {
	          ss << "x";
		  ss << axisLengths[zz]; 
	       }
               ss >> ext_p->dim;
	    }
            free(axisLengths);
         }
	 ext_p++; 
      }
      // reopen the opened extension 
      if (open_block != -1)  open_extension(open_block);  
   }
   return 0; 
}

string PrismDataSet::get_current_extname()
{
   static string ename;
   if (_block)   ename = _block->get_name(); 
      else       ename = "";  
   return ename;    
}

int PrismDataSet::set_editmode()
{
   if (_editmode == 1) // locked - set mode to edit or modified 
   {
      if (_firstunlock)
      {
         // query user 
	 PrismBackupDialog *dlg = new PrismBackupDialog("Edit Confirmation", _fname); 
         if (dlg->doit() != 0)
	 {
	    return 1; 
	 } 	 
         _firstunlock = false; 
      }
      if (_modified)
         _editmode = 3;
      else
         _editmode = 2; 
   }
   else if (_editmode == 2 || _editmode == 3)  // edit or modified- set to locked
   {
      _editmode = 1; 
   }
   // for r/o and default- do nothing since the mode doesn't change 
   return _editmode; 
}

void PrismDataSet::set_modified(int val)
{
   _modified = val; 
   if (_editmode = 2)  _editmode = 3; 
   if (_editmode == 3)  flush_buffer();
}   
    
void PrismDataSet::flush_buffer()
{   
   PrismDataBlock* pdb = get_block();
   if (pdb)  pdb->flush_changes();
}
