// 
//  Copyright (C) 2003,2007-2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


//     This file contains the source code for the CiaoToolEntry class. The
//     class is used to store command / menu components for the Analysis
//     Pulldown menu. This class categorizes analysis menu descriptor 
//     entries into one of for categories: Category (menu pulldown label),
//     Command (a task to execute), Separator (separator line in the menu),
//     or Help (information to display in a help window). In addition, 
//     this class also supports functionality for the categories. For
//     instance, macro expansion in commands.

#include <gtk/gtk.h>
#include <stdlib.h>
#include "CiaoAnalysisMenu.hh"
#include "CiaoToolEntry.hh"
#include "CiaoTextEntryDialog.hh"
#include "CiaoGuiUtils.hh"

// constants
const int CiaoToolEntry::CATEGORY_TYPE   = 0;
const int CiaoToolEntry::COMMAND_TYPE    = 1;
const int CiaoToolEntry::SEPARATOR_TYPE  = 2;
const int CiaoToolEntry::HELP_TYPE       = 3;
const int CiaoToolEntry::SUPPORTED       = 0;
const int CiaoToolEntry::NOT_SUPPORTED   = 1;
const int CiaoToolEntry::NEVER_SUPPORTED = 2;


// routine to do a case insensitive substring search of a string
static char* strstrLoose(const char *src, const char *sub)
{
   int srclen = strlen(src);
   int sublen = strlen(sub);
   int rr = 0, cc = 0, match = 0, curr = tolower(sub[0]);
   char *pos = NULL;

   while ((rr < srclen) && (match < sublen))
   {
      if (tolower(src[rr]) == curr)
      {
         if (cc == 0)   pos = (char*)&src[rr];
         ++match;
         curr = tolower(sub[++cc]);
      }
      else
      {
         match = cc = 0;
         curr = tolower(sub[0]);
      }
      ++rr;
   }
   if (match != sublen)   pos = NULL;

   return (pos);
}


// Constructor: Separator
CiaoToolEntry::CiaoToolEntry(int depth)   // menu depth 
{
   static int scnt = 1;
   char buff[256];

   sprintf(buff, "sep%1d", scnt++); 
   _name = buff;
   _cmd = "";
   _template = "";
   _cmdType = "";
   _tip = "";
   _depth = depth; 
   _type = SEPARATOR_TYPE;
   _supported = SUPPORTED; 
   _replace_filename = NULL; 
   _status_message = NULL; 
}

// Constructor: Help
CiaoToolEntry::CiaoToolEntry(
   string name,   // menu label  
   string txt,    // help text  
   int format,   // format options  
   int depth)    // menu depth 
{
   _name = name;
   _help = txt; 
   _supported = SUPPORTED;
   _depth = depth;
   _type = HELP_TYPE;
   _replace_filename = NULL;
   _status_message = NULL;
}


// Constructor: Category 
CiaoToolEntry::CiaoToolEntry(
   string name,  // menu label  
   string tip,   // tool tip  
   int depth)   // menu depth 
{
   _name = name; 
   _cmd = "";
   _template = "";
   _cmdType = "menu"; 
   _tip = tip; 
   _supported = SUPPORTED; 
   _depth = depth;
   _type = CATEGORY_TYPE; 
   _replace_filename = NULL;
   _status_message = NULL;
}


CiaoToolEntry::CiaoToolEntry(
   string name,      // label for entry  
   string tip,       // tool tip  
   string ftype,     // supported file types  
   string ctype,     // command type -entry/bind  
   string cmd,       // actual command   
   int depth)       // menu depth 
{
   _supported = NEVER_SUPPORTED; 
   _name = name;
   _tip = tip;
   _template = ftype;
   if (_template.length())   _formats = _ciao_string_split(_template); 
   _cmdType = ctype; 
   if (_ciao_strcmp_loose(_cmdType,"menu") == 0)  _supported = SUPPORTED;
   _cmd = cmd;
   _depth = depth;
   _type = COMMAND_TYPE;
   _replace_filename = NULL;
   _status_message = NULL;
}


CiaoToolEntry::~CiaoToolEntry() 
{
}


void CiaoToolEntry::doit()
{
   if (_supported == SUPPORTED)
   {
      if (_type == COMMAND_TYPE)        // command 
      {
         ReconfigureCmd(); 

         // expand macros and run tool
         ExpandMacros(_cmdEval); 
      } 
      else if (_type == HELP_TYPE)   // help 
      {
         string txt = _name; 
         txt.append("\n\n");
         txt.append(_help);
         GtkWidget* msg_dialog = gtk_message_dialog_new(NULL, 
            GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, 
	    GTK_BUTTONS_OK, txt.c_str());
	 gtk_widget_show_all(msg_dialog);
         if (gtk_dialog_run(GTK_DIALOG(msg_dialog)) == GTK_RESPONSE_OK)
            gtk_widget_destroy(msg_dialog);
      } 
   } 
}

int CiaoToolEntry::DisableUnsupported()
{
   static char* unsupported[] = {
     "$data", "$xdim", "$ydim", "$bitpix", "$x", "$y", "$plot",  
     "$url", "$geturl", "$xpa", "$xpamethod", "$param", "$regions" };
   static int unlen[] = {5,5,5,7,2,2,5,4,7,4,10,6,8};
   int unsize = 13; 
   string::size_type cmdlen = _cmd.length(); 
   int invalid = 0;
   
   if (cmdlen > 1)
   {
      int rr;
      char* pos; 
      char* temp = new char[cmdlen + 1];

      if (temp)
      {
         strcpy(temp, _cmd.c_str());

         rr = unsize;
         while (!invalid && rr--)
         {
            if ((pos = strstrLoose(temp, unsupported[rr])) != NULL)
            {
               // check that macro is not a substring ie. $x vs. $xray
               char *nxt = pos + unlen[rr]; 
               if (!isalnum(*nxt))  invalid = 1;  
            }
         }   // end for(unsize)

         if (invalid) 
         {
            _supported = NOT_SUPPORTED;
            return 1; 
         }
         delete []temp;
      }
   }
   return 0; 
}

void CiaoToolEntry::SupportByFormat(string type)
{
   _supTemp = NOT_SUPPORTED; 
   if (_type == COMMAND_TYPE && (_supported == SUPPORTED))
   {
      vector <string>::iterator it;
      int found = 0;

      for (it =_formats.begin(); it != _formats.end(); it++)
         found = ((_ciao_strcmp_cis(*it,type)== 0) || (*it == "*")); 
      
      if (found) 
      {
         _supTemp = SUPPORTED; 
//WLM         if (_supported == SUPPORTED)   activate();
      }
      else 
      {
//WLM         if (_supported == SUPPORTED)   deactivate();
      }
   }
}


// convert to non ds9 specific format 
void CiaoToolEntry::ReconfigureCmd()
{
   string::size_type cmdlen = _cmd.length();
   if (cmdlen > 1)
   {
      char* pos;
      char* temp = new char[cmdlen + 1];

      _cmdEval.clear(); 
      if (temp)
      {
         strcpy(temp, _cmd.c_str());

         // check for $text  
         if ((pos = strstrLoose(temp, "| $text")) != NULL)
         {
            // remove $text 
            *pos = '\0'; 

            // if taskmonitor or peg specified- all set
            if ((strstrLoose(temp, "taskmonitor") == NULL) &&
               (strstrLoose(temp, "peg") == NULL)) 
            {
               _cmdEval = "taskmonitor \"";
               _cmdEval.append(temp); 
               _cmdEval.append("\""); 
            }
            else 
            {
               _cmdEval.append(temp); 
            }
         }
         else if ((pos = strstrLoose(temp, "| $null")) != NULL)
         {
            // remove $null macro
            *pos = '\0';
            _cmdEval = temp; 
         }

         if (!_cmdEval.length())
         {
            _cmdEval = _cmd; 
         } 

         delete []temp;
      }
   } 
}


void CiaoToolEntry::ExpandMacros(string val)
{
   string sval = val;
   vector <string> mylist = _ciao_string_split(sval, "$");

   if (mylist.size() > 0)
   {
      vector <string>::iterator it;
      int cont = 1; 

      _cmdEval.clear(); 
      for(it = mylist.begin(); it != mylist.end(); it++)
      {
         string lab, current = *it;
         string::size_type clen = current.length();
         if (clen > 1)
         {
            char *temp = new char[clen + 1];

            if (temp) 
            {
               strcpy(temp, current.c_str());
         
               // check for $text
               if (strstrLoose(temp, "message") == temp)
               {
                  GtkButtonsType btns_type;

                  // determine message type- ok | okcancel | yesno
                  char *ps = strchr(temp,'(');
                  if (ps)
                  {
                     ++ps;
                     while (isspace(*ps)) ps++;  
                     if (strstrLoose(ps, "okcancel") == ps) 
		        btns_type = GTK_BUTTONS_OK_CANCEL;
                     else if (strstrLoose(ps, "yesno") == ps) 
		        btns_type = GTK_BUTTONS_YES_NO;
                     else
		        btns_type = GTK_BUTTONS_OK; 
		     lab = ps; 
                  }
                  if ((ps = strchr(temp,',')) != NULL)   lab = ++ps; 
		  lab = lab.substr(0, lab.find(")")); 

                  if ((ps = strchr(temp,'|')) != NULL)
                     _cmdEval.append(++ps); 

                  GtkWidget* msg_dialog = gtk_message_dialog_new(NULL, 
                     GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, 
	             btns_type, lab.c_str());
	          gtk_widget_show_all(msg_dialog);
		  int resp = gtk_dialog_run(GTK_DIALOG(msg_dialog));
		  if (resp == GTK_RESPONSE_NO || resp == GTK_RESPONSE_CANCEL)
                     cont = 0; 
                  gtk_widget_destroy(msg_dialog);
               }
               else if (strstrLoose(temp, "entry") == temp)
               {
                  string::size_type beg = current.find('('),  
                     end = current.find(')'); 

                  if ((beg != string::npos) && (end != string::npos) && 
                      (beg < end))
                     lab = current.substr(beg+1, end-beg-1); 
		  else
		     lab.clear(); 
                  string rest = current.substr(end+1, clen); 
		  rest = _ciao_substring_replace(rest, "|", " ");  

                 CiaoTextEntryDialog *ted = new CiaoTextEntryDialog(lab);
                 string val;
		 if (ted && !ted->get_value(val))
                 {
		    _cmdEval.append(val);
		    _cmdEval.append(rest.c_str());
		 }
                 else cont = 0; // hit cancel so don't execute command 
               }
               else if (strstrLoose(temp, "filename(root)") == temp)
               {
                  string fname, rname="", cval = *it;
                  string old = current.substr(0, 13);
		  cval = _ciao_substring_replace(cval, old, rname);
                  _cmdEval.append(cval);
               }
               else if (strstrLoose(temp, "filename") == temp)
               {
                  string fname = "", cval = *it;
		  if (_replace_filename) fname = _replace_filename(); 
                  if (fname.find(" ") != string::npos) 
                  {
                     // if the filename has a filter containing a space
                     // enclose in single quotes to avoid parsing issues later
                     if (fname.find("'") == string::npos)
                     {    
                        fname = "'" + fname + "'";
                     }
                  }
                  string old = current.substr(0, 8); 
		  cval = _ciao_substring_replace(cval, old, fname); 
                  _cmdEval.append(cval); 
               }
               else 
                  _cmdEval.append(temp);
               
               delete []temp;
            } 
         } 
      } 
      if (cont)
      {
         // execute the task 
         _cmdEval.append(" &");
	 if (_status_message) 
	 {
	    string msg = "Executing Analysis Menu command: " + _name;  
	    _status_message(msg); 
	 }
         system(_cmdEval.c_str()); 
      }
   }
}


